from django.test import TestCase
from django.db import IntegrityError
from django.core.exceptions import ValidationError

from address.models import Country, State, Locality, Address, AddressField
from address.models import to_python

# Python 3 fixes.
import sys

if sys.version > "3":
    long = int
    basestring = (str, bytes)
    unicode = str


class CountryTestCase(TestCase):
    def setUp(self):
        self.au = Country.objects.create(name="Australia", code="AU")
        self.nz = Country.objects.create(name="New Zealand", code="NZ")
        self.be = Country.objects.create(name="Belgium", code="BE")

    def test_ordering(self):
        qs = Country.objects.all()
        self.assertEqual(qs.count(), 3)
        self.assertEqual(qs[0].code, "AU")
        self.assertEqual(qs[1].code, "BE")
        self.assertEqual(qs[2].code, "NZ")

    def test_unique_name(self):
        self.assertRaises(IntegrityError, Country.objects.create, name="Australia", code="**")

    def test_unicode(self):
        self.assertEqual(unicode(self.au), u"Australia")


class StateTestCase(TestCase):
    def setUp(self):
        self.au = Country.objects.create(name="Australia", code="AU")
        self.vic = State.objects.create(name="Victoria", code="VIC", country=self.au)
        self.tas = State.objects.create(name="Tasmania", code="TAS", country=self.au)
        self.qld = State.objects.create(name="Queensland", country=self.au)
        self.empty = State.objects.create(country=self.au)
        self.uk = Country.objects.create(name="United Kingdom", code="UK")
        self.uk_vic = State.objects.create(name="Victoria", code="VIC", country=self.uk)

    def test_required_country(self):
        self.assertRaises(IntegrityError, State.objects.create)

    def test_ordering(self):
        qs = State.objects.all()
        self.assertEqual(qs.count(), 5)
        self.assertEqual(qs[0].name, "")
        self.assertEqual(qs[1].name, "Queensland")
        self.assertEqual(qs[2].name, "Tasmania")
        self.assertEqual(qs[3].name, "Victoria")
        self.assertEqual(qs[4].name, "Victoria")

    def test_unique_name_country(self):
        State.objects.create(name="Tasmania", country=self.uk)
        self.assertRaises(IntegrityError, State.objects.create, name="Tasmania", country=self.au)

    def test_unicode(self):
        self.assertEqual(unicode(self.vic), u"Victoria, Australia")
        self.assertEqual(unicode(self.empty), u"Australia")


class LocalityTestCase(TestCase):
    def setUp(self):
        self.au = Country.objects.create(name="Australia", code="AU")
        self.uk = Country.objects.create(name="United Kingdom", code="UK")

        self.au_vic = State.objects.create(name="Victoria", code="VIC", country=self.au)
        self.au_tas = State.objects.create(name="Tasmania", code="TAS", country=self.au)
        self.au_qld = State.objects.create(name="Queensland", country=self.au)
        self.au_empty = State.objects.create(country=self.au)
        self.uk_vic = State.objects.create(name="Victoria", code="VIC", country=self.uk)

        self.au_vic_nco = Locality.objects.create(name="Northcote", postal_code="3070", state=self.au_vic)
        self.au_vic_mel = Locality.objects.create(name="Melbourne", postal_code="3000", state=self.au_vic)
        self.au_vic_ftz = Locality.objects.create(name="Fitzroy", state=self.au_vic)
        self.au_vic_empty = Locality.objects.create(state=self.au_vic)
        self.uk_vic_mel = Locality.objects.create(name="Melbourne", postal_code="3000", state=self.uk_vic)

    def test_required_state(self):
        self.assertRaises(IntegrityError, Locality.objects.create)

    def test_ordering(self):
        qs = Locality.objects.all()
        self.assertEqual(qs.count(), 5)
        self.assertEqual(qs[0].name, "")
        self.assertEqual(qs[1].name, "Fitzroy")
        self.assertEqual(qs[2].name, "Melbourne")
        self.assertEqual(qs[3].name, "Northcote")
        self.assertEqual(qs[4].name, "Melbourne")

    def test_unicode(self):
        self.assertEqual(unicode(self.au_vic_mel), u"Melbourne, Victoria 3000, Australia")
        self.assertEqual(unicode(self.au_vic_ftz), u"Fitzroy, Victoria, Australia")
        self.assertEqual(unicode(self.au_vic_empty), u"Victoria, Australia")


class AddressTestCase(TestCase):
    def setUp(self):
        self.au = Country.objects.create(name="Australia", code="AU")
        self.uk = Country.objects.create(name="United Kingdom", code="UK")

        self.au_vic = State.objects.create(name="Victoria", code="VIC", country=self.au)
        self.au_tas = State.objects.create(name="Tasmania", code="TAS", country=self.au)
        self.au_qld = State.objects.create(name="Queensland", country=self.au)
        self.au_empty = State.objects.create(country=self.au)
        self.uk_vic = State.objects.create(name="Victoria", code="VIC", country=self.uk)

        self.au_vic_nco = Locality.objects.create(name="Northcote", postal_code="3070", state=self.au_vic)
        self.au_vic_mel = Locality.objects.create(name="Melbourne", postal_code="3000", state=self.au_vic)
        self.au_vic_ftz = Locality.objects.create(name="Fitzroy", state=self.au_vic)
        self.au_vic_empty = Locality.objects.create(state=self.au_vic)
        self.uk_vic_mel = Locality.objects.create(name="Melbourne", postal_code="3000", state=self.uk_vic)

        self.ad1 = Address.objects.create(
            street_number="1",
            route="Some Street",
            locality=self.au_vic_mel,
            raw="1 Some Street, Victoria, Melbourne",
        )
        self.ad2 = Address.objects.create(
            street_number="10",
            route="Other Street",
            locality=self.au_vic_mel,
            raw="10 Other Street, Victoria, Melbourne",
        )
        self.ad3 = Address.objects.create(
            street_number="1",
            route="Some Street",
            locality=self.au_vic_nco,
            raw="1 Some Street, Northcote, Victoria",
        )
        self.ad_empty = Address.objects.create(locality=self.au_vic_nco, raw="Northcote, Victoria")

    def test_required_raw(self):
        obj = Address.objects.create()
        self.assertRaises(ValidationError, obj.clean)

    def test_ordering(self):
        qs = Address.objects.all()
        self.assertEqual(qs.count(), 4)
        self.assertEqual(qs[0].route, "Other Street")
        self.assertEqual(qs[1].route, "Some Street")
        self.assertEqual(qs[2].route, "")
        self.assertEqual(qs[3].route, "Some Street")

    # def test_unique_street_address_locality(self):
    #     Address.objects.create(street_number='10', route='Other Street', locality=self.au_vic_nco)
    #     self.assertRaises(
    #         IntegrityError, Address.objects.create,
    #         street_number='10', route='Other Street', locality=self.au_vic_mel
    #     )

    def test_unicode(self):
        self.assertEqual(unicode(self.ad1), u"1 Some Street, Melbourne, Victoria 3000, Australia")
        self.assertEqual(unicode(self.ad_empty), u"Northcote, Victoria 3070, Australia")


class AddressFieldTestCase(TestCase):
    class TestModel(object):
        address = AddressField()

    def setUp(self):
        self.ad1_dict = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "locality": "Northcote",
            "postal_code": "3070",
            "state": "Victoria",
            "state_code": "VIC",
            "country": "Australia",
            "country_code": "AU",
        }
        self.test = self.TestModel()

    def test_assignment_from_dict(self):
        self.test.address = to_python(self.ad1_dict)
        self.assertEqual(self.test.address.raw, self.ad1_dict["raw"])
        self.assertEqual(self.test.address.street_number, self.ad1_dict["street_number"])
        self.assertEqual(self.test.address.route, self.ad1_dict["route"])
        self.assertEqual(self.test.address.locality.name, self.ad1_dict["locality"])
        self.assertEqual(self.test.address.locality.postal_code, self.ad1_dict["postal_code"])
        self.assertEqual(self.test.address.locality.state.name, self.ad1_dict["state"])
        self.assertEqual(self.test.address.locality.state.code, self.ad1_dict["state_code"])
        self.assertEqual(self.test.address.locality.state.country.name, self.ad1_dict["country"])
        self.assertEqual(self.test.address.locality.state.country.code, self.ad1_dict["country_code"])

    def test_assignment_from_dict_no_country(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "locality": "Northcote",
            "state": "Victoria",
        }
        self.test.address = to_python(ad)
        self.assertEqual(self.test.address.raw, ad["raw"])
        self.assertEqual(self.test.address.street_number, "")
        self.assertEqual(self.test.address.route, "")
        self.assertEqual(self.test.address.locality, None)

    def test_assignment_from_dict_no_state(self):
        ad = {
            "raw": "Somewhere",
            "locality": "Northcote",
            "country": "Australia",
        }
        self.test.address = to_python(ad)
        self.assertEqual(self.test.address.raw, ad["raw"])
        self.assertEqual(self.test.address.street_number, "")
        self.assertEqual(self.test.address.route, "")
        self.assertEqual(self.test.address.locality, None)

    def test_assignment_from_dict_no_locality(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "state": "Victoria",
            "country": "Australia",
        }
        self.test.address = to_python(ad)
        self.assertEqual(self.test.address.raw, ad["raw"])
        self.assertEqual(self.test.address.street_number, "")
        self.assertEqual(self.test.address.route, "")
        self.assertEqual(self.test.address.locality, None)

    def test_assignment_from_dict_only_address(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
        }
        self.test.address = to_python(ad)
        self.assertEqual(self.test.address.raw, ad["raw"])
        self.assertEqual(self.test.address.street_number, ad["street_number"])
        self.assertEqual(self.test.address.route, ad["route"])
        self.assertEqual(self.test.address.locality, None)

    def test_assignment_from_dict_duplicate_country_code(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "locality": "Northcote",
            "state": "Victoria",
            "country": "Australia",
            "country_code": "Australia",
        }
        self.test.address = to_python(ad)
        self.assertEqual(self.test.address.raw, ad["raw"])
        self.assertEqual(self.test.address.street_number, "1")
        self.assertEqual(self.test.address.route, "Somewhere Street")
        self.assertEqual(self.test.address.locality.name, "Northcote")
        self.assertEqual(self.test.address.locality.state.name, "Victoria")
        self.assertEqual(self.test.address.locality.state.country.name, "Australia")
        self.assertEqual(self.test.address.locality.state.country.code, "")

    def test_assignment_from_dict_duplicate_state_code(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "locality": "Northcote",
            "state": "Victoria",
            "state_code": "Victoria",
            "country": "Australia",
        }
        self.test.address = to_python(ad)
        self.assertEqual(self.test.address.raw, ad["raw"])
        self.assertEqual(self.test.address.street_number, "1")
        self.assertEqual(self.test.address.route, "Somewhere Street")
        self.assertEqual(self.test.address.locality.name, "Northcote")
        self.assertEqual(self.test.address.locality.state.name, "Victoria")
        self.assertEqual(self.test.address.locality.state.code, "Victoria")
        self.assertEqual(self.test.address.locality.state.country.name, "Australia")

    def test_assignment_from_dict_invalid_country_code(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "locality": "Northcote",
            "state": "Victoria",
            "country": "Australia",
            "country_code": "Something else",
        }
        self.assertRaises(ValueError, to_python, ad)

    def test_assignment_from_dict_invalid_state_code(self):
        ad = {
            "raw": "1 Somewhere Street, Northcote, Victoria 3070, VIC, AU",
            "street_number": "1",
            "route": "Somewhere Street",
            "locality": "Northcote",
            "state": "Victoria",
            "state_code": "Something",
            "country": "Australia",
        }
        # This is invalid because state codes are expected to have a max of 8 characters
        self.assertRaises(ValueError, to_python, ad)

    def test_assignment_from_string(self):
        self.test.address = to_python(self.ad1_dict["raw"])
        self.assertEqual(self.test.address.raw, self.ad1_dict["raw"])

    # def test_save(self):
    #     self.test.address = self.ad1_dict
    #     self.test.save()
    #     test = self.TestModel.objects.all()[0]
    #     self.assertEqual(test.address.raw, self.ad1_dict['raw'])
    #     self.assertEqual(test.address.street_number, self.ad1_dict['street_number'])
    #     self.assertEqual(test.address.route, self.ad1_dict['route'])
    #     self.assertEqual(test.address.locality.name, self.ad1_dict['locality'])
    #     self.assertEqual(test.address.locality.postal_code, self.ad1_dict['postal_code'])
    #     self.assertEqual(test.address.locality.state.name, self.ad1_dict['state'])
    #     self.assertEqual(test.address.locality.state.code, self.ad1_dict['state_code'])
    #     self.assertEqual(test.address.locality.state.country.name, self.ad1_dict['country'])
    #     self.assertEqual(test.address.locality.state.country.code, self.ad1_dict['country_code'])



from django.test import TestCase
from django.forms import ValidationError, Form
from address.forms import AddressField, AddressWidget


class TestForm(Form):
    address = AddressField()


class AddressFieldTestCase(TestCase):
    def setUp(self):
        self.form = TestForm()
        self.field = self.form.base_fields["address"]
        self.missing_state = {
            "country": "UK",
            "locality": "Somewhere",
            "postal_code": "34904",
            "route": "A street?",
            "street_number": "3",
            "raw": "3 A street?, Somewhere, UK",
        }

    def test_to_python_none(self):
        self.assertEqual(self.field.to_python(None), None)

    def test_to_python_empty(self):
        self.assertEqual(self.field.to_python(""), None)

    def test_to_python_invalid_lat_lng(self):
        self.assertRaises(ValidationError, self.field.to_python, {"latitude": "x"})
        self.assertRaises(ValidationError, self.field.to_python, {"longitude": "x"})

    def test_to_python_invalid_empty_lat_lng(self):
        self.assertEqual(self.field.to_python({"latitude": ""}), None)
        self.assertEqual(self.field.to_python({"longitude": ""}), None)

    def test_to_python_no_locality(self):
        input = {
            "country": "United States",
            "country_code": "US",
            "state": "New York",
            "state_code": "NY",
            "locality": "",
            "sublocality": "Brooklyn",
            "postal_code": "11201",
            "route": "Joralemon St",
            "street_number": "209",
            "raw": "209 Joralemon Street, Brooklyn, NY, United States",
        }
        res = self.field.to_python(input)
        self.assertEqual(res.locality.name, "Brooklyn")

    def test_to_python_postal_town(self):
        """UK addresses with no `locality`, but a populated `postal_town`, should use the
        `postal_town` as the `locality`"""
        data = {
            "raw": "High Street, Leamington Spa",
            "route": "High Street",
            "postal_town": "Leamington Spa",
            "state": "England",
            "state_code": "England",
            "country": "United Kingdom",
            "country_code": "GB",
            "postal_code": "CV31",
            "formatted": "High St, Royal Leamington Spa, Leamington Spa CV31, UK",
        }
        address = self.field.to_python(data)
        self.assertIsNotNone(address.locality)
        self.assertEqual(address.locality.name, data["postal_town"])

    # TODO: Fix
    # def test_to_python_empty_state(self):
    #     val = self.field.to_python(self.missing_state)
    #     self.assertTrue(isinstance(val, Address))
    #     self.assertNotEqual(val.locality, None)

    def test_to_python(self):
        res = self.field.to_python({"raw": "Someplace"})
        self.assertEqual(res.raw, "Someplace")

    def test_render(self):
        # TODO: Check return value.
        self.form.as_table()


class AddressWidgetTestCase(TestCase):
    def test_attributes_set_correctly(self):
        wid = AddressWidget(attrs={"size": "150"})
        self.assertEqual(wid.attrs["size"], "150")
        html = wid.render("test", None)
        self.assertNotEqual(html.find('size="150"'), -1)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ("address", "0001_initial"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="locality",
            unique_together=set([("name", "postal_code", "state")]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Address",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("street_number", models.CharField(max_length=20, blank=True)),
                ("route", models.CharField(max_length=100, blank=True)),
                ("raw", models.CharField(max_length=200)),
                ("formatted", models.CharField(max_length=200, blank=True)),
                ("latitude", models.FloatField(null=True, blank=True)),
                ("longitude", models.FloatField(null=True, blank=True)),
            ],
            options={
                "ordering": ("locality", "route", "street_number"),
                "verbose_name_plural": "Addresses",
            },
        ),
        migrations.CreateModel(
            name="Country",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(unique=True, max_length=40, blank=True)),
                ("code", models.CharField(max_length=2, blank=True)),
            ],
            options={
                "ordering": ("name",),
                "verbose_name_plural": "Countries",
            },
        ),
        migrations.CreateModel(
            name="Locality",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=165, blank=True)),
                ("postal_code", models.CharField(max_length=10, blank=True)),
            ],
            options={
                "ordering": ("state", "name"),
                "verbose_name_plural": "Localities",
            },
        ),
        migrations.CreateModel(
            name="State",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=165, blank=True)),
                ("code", models.CharField(max_length=3, blank=True)),
                (
                    "country",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        related_name="states",
                        to="address.Country",
                    ),
                ),
            ],
            options={
                "ordering": ("country", "name"),
            },
        ),
        migrations.AddField(
            model_name="locality",
            name="state",
            field=models.ForeignKey(on_delete=models.CASCADE, related_name="localities", to="address.State"),
        ),
        migrations.AddField(
            model_name="address",
            name="locality",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="addresses",
                blank=True,
                to="address.Locality",
                null=True,
            ),
        ),
        migrations.AlterUniqueTogether(
            name="state",
            unique_together=set([("name", "country")]),
        ),
        migrations.AlterUniqueTogether(
            name="locality",
            unique_together=set([("name", "state")]),
        ),
    ]



from django.contrib import admin
from django.contrib.admin import SimpleListFilter

from address.models import Country, State, Locality, Address


class UnidentifiedListFilter(SimpleListFilter):
    title = "unidentified"
    parameter_name = "unidentified"

    def lookups(self, request, model_admin):
        return (("unidentified", "unidentified"),)

    def queryset(self, request, queryset):
        if self.value() == "unidentified":
            return queryset.filter(locality=None)


@admin.register(Country)
class CountryAdmin(admin.ModelAdmin):
    search_fields = ("name", "code")


@admin.register(State)
class StateAdmin(admin.ModelAdmin):
    search_fields = ("name", "code")


@admin.register(Locality)
class LocalityAdmin(admin.ModelAdmin):
    search_fields = ("name", "postal_code")


@admin.register(Address)
class AddressAdmin(admin.ModelAdmin):
    search_fields = ("street_number", "route", "raw")
    list_filter = (UnidentifiedListFilter,)



import logging

from django import forms

from .models import Address, to_python
from .widgets import AddressWidget

logger = logging.getLogger(__name__)

__all__ = ["AddressWidget", "AddressField"]


class AddressField(forms.ModelChoiceField):
    widget = AddressWidget

    def __init__(self, *args, **kwargs):
        kwargs["queryset"] = Address.objects.none()
        super(AddressField, self).__init__(*args, **kwargs)

    def to_python(self, value):

        # Treat `None`s and empty strings as empty.
        if value is None or value == "":
            return None

        # Check for garbage in the lat/lng components.
        for field in ["latitude", "longitude"]:
            if field in value:
                if value[field]:
                    try:
                        value[field] = float(value[field])
                    except Exception:
                        raise forms.ValidationError(
                            "Invalid value for %(field)s",
                            code="invalid",
                            params={"field": field},
                        )
                else:
                    value[field] = None

        return to_python(value)



from django.apps import AppConfig


class AddressConfig(AppConfig):
    """
    Define config for the member app so that we can hook in signals.
    """

    name = "address"
    default_auto_field = "django.db.models.AutoField"



import django
from django import forms
from django.conf import settings
from django.utils.html import escape
from django.utils.safestring import mark_safe

from .models import Address

USE_DJANGO_JQUERY = getattr(settings, "USE_DJANGO_JQUERY", False)
JQUERY_URL = getattr(
    settings,
    "JQUERY_URL",
    "https://ajax.googleapis.com/ajax/libs/jquery/2.2.0/jquery.min.js",
)


class AddressWidget(forms.TextInput):
    components = [
        ("country", "country"),
        ("country_code", "country_short"),
        ("locality", "locality"),
        ("sublocality", "sublocality"),
        ("postal_code", "postal_code"),
        ("postal_town", "postal_town"),
        ("route", "route"),
        ("street_number", "street_number"),
        ("state", "administrative_area_level_1"),
        ("state_code", "administrative_area_level_1_short"),
        ("formatted", "formatted_address"),
        ("latitude", "lat"),
        ("longitude", "lng"),
    ]

    class Media:
        """Media defined as a dynamic property instead of an inner class."""

        js = [
            "https://maps.googleapis.com/maps/api/js?libraries=places&key=%s" % settings.GOOGLE_API_KEY,
            "js/jquery.geocomplete.min.js",
            "address/js/address.js",
        ]

        if JQUERY_URL:
            js.insert(0, JQUERY_URL)
        elif JQUERY_URL is not False:
            vendor = "" if django.VERSION < (1, 9, 0) else "vendor/jquery/"
            extra = "" if settings.DEBUG else ".min"

            jquery_paths = [
                "{}jquery{}.js".format(vendor, extra),
                "jquery.init.js",
            ]

            if USE_DJANGO_JQUERY:
                jquery_paths = ["admin/js/{}".format(path) for path in jquery_paths]

            js.extend(jquery_paths)

    def __init__(self, *args, **kwargs):
        attrs = kwargs.get("attrs", {})
        classes = attrs.get("class", "")
        classes += (" " if classes else "") + "address"
        attrs["class"] = classes
        kwargs["attrs"] = attrs
        super(AddressWidget, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs=None, **kwargs):

        # Can accept None, a dictionary of values or an Address object.
        if value in (None, ""):
            ad = {}
        elif isinstance(value, dict):
            ad = value
        elif isinstance(value, int):
            ad = Address.objects.get(pk=value)
            ad = ad.as_dict()
        else:
            ad = value.as_dict()

        # Generate the elements. We should create a suite of hidden fields
        # For each individual component, and a visible field for the raw
        # input. Begin by generating the raw input.
        elems = [super(AddressWidget, self).render(name, escape(ad.get("formatted", "")), attrs, **kwargs)]

        # Now add the hidden fields.
        elems.append('<div id="%s_components" style="display: none;">' % name)
        for com in self.components:
            elems.append(
                '<input type="hidden" name="%s_%s" data-geo="%s" value="%s" />'
                % (name, com[0], com[1], escape(ad.get(com[0], "")))
            )
        elems.append("</div>")

        return mark_safe("\n".join(elems))

    def value_from_datadict(self, data, files, name):
        raw = data.get(name, "")
        if not raw:
            return raw
        ad = dict([(c[0], data.get(name + "_" + c[0], "")) for c in self.components])
        ad["raw"] = raw
        return ad



import os
import sys

from setuptools import find_packages, setup

version = "0.2.8"

if sys.argv[-1] == "tag":
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

setup(
    name="django-address",
    version=version,
    author="Luke Hodkinson",
    author_email="furious.luke@gmail.com",
    maintainer="Rob Banagale",
    maintainer_email="rob@banagale.com",
    url="https://github.com/furious-luke/django-address",
    description="A django application for describing addresses.",
    long_description=open(os.path.join(os.path.dirname(__file__), "README.md")).read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    license="BSD",
    packages=find_packages(),
    include_package_data=True,
    package_data={"": ["*.txt", "*.js", "*.html", "*.*"]},
    install_requires=["setuptools"],
    zip_safe=False,
)



import logging

from django.core.exceptions import ValidationError
from django.db import models

try:
    from django.db.models.fields.related_descriptors import ForwardManyToOneDescriptor
except ImportError:
    from django.db.models.fields.related import (
        ReverseSingleRelatedObjectDescriptor as ForwardManyToOneDescriptor,
    )

logger = logging.getLogger(__name__)

__all__ = ["Country", "State", "Locality", "Address", "AddressField"]


class InconsistentDictError(Exception):
    pass


def _to_python(value):
    raw = value.get("raw", "")
    country = value.get("country", "")
    country_code = value.get("country_code", "")
    state = value.get("state", "")
    state_code = value.get("state_code", "")
    locality = value.get("locality", "")
    sublocality = value.get("sublocality", "")
    postal_town = value.get("postal_town", "")
    postal_code = value.get("postal_code", "")
    street_number = value.get("street_number", "")
    route = value.get("route", "")
    formatted = value.get("formatted", "")
    latitude = value.get("latitude", None)
    longitude = value.get("longitude", None)

    # If there is no value (empty raw) then return None.
    if not raw:
        return None

    # Fix issue with NYC boroughs (https://code.google.com/p/gmaps-api-issues/issues/detail?id=635)
    if not locality and sublocality:
        locality = sublocality

    # Fix issue with UK addresses with no locality
    # (https://github.com/furious-luke/django-address/issues/114)
    if not locality and postal_town:
        locality = postal_town

    # If we have an inconsistent set of value bail out now.
    if (country or state or locality) and not (country and state and locality):
        raise InconsistentDictError

    # Handle the country.
    try:
        country_obj = Country.objects.get(name=country)
    except Country.DoesNotExist:
        if country:
            if len(country_code) > Country._meta.get_field("code").max_length:
                if country_code != country:
                    raise ValueError("Invalid country code (too long): %s" % country_code)
                country_code = ""
            country_obj = Country.objects.create(name=country, code=country_code)
        else:
            country_obj = None

    # Handle the state.
    try:
        state_obj = State.objects.get(name=state, country=country_obj)
    except State.DoesNotExist:
        if state:
            if len(state_code) > State._meta.get_field("code").max_length:
                if state_code != state:
                    raise ValueError("Invalid state code (too long): %s" % state_code)
                state_code = ""
            state_obj = State.objects.create(name=state, code=state_code, country=country_obj)
        else:
            state_obj = None

    # Handle the locality.
    try:
        locality_obj = Locality.objects.get(name=locality, postal_code=postal_code, state=state_obj)
    except Locality.DoesNotExist:
        if locality:
            locality_obj = Locality.objects.create(name=locality, postal_code=postal_code, state=state_obj)
        else:
            locality_obj = None

    # Handle the address.
    try:
        if not (street_number or route or locality):
            address_obj = Address.objects.get(raw=raw)
        else:
            address_obj = Address.objects.get(street_number=street_number, route=route, locality=locality_obj)
    except Address.DoesNotExist:
        address_obj = Address(
            street_number=street_number,
            route=route,
            raw=raw,
            locality=locality_obj,
            formatted=formatted,
            latitude=latitude,
            longitude=longitude,
        )

        # If "formatted" is empty try to construct it from other values.
        if not address_obj.formatted:
            address_obj.formatted = str(address_obj)

        # Need to save.
        address_obj.save()

    # Done.
    return address_obj


##
# Convert a dictionary to an address.
##


def to_python(value):

    # Keep `None`s.
    if value is None:
        return None

    # Is it already an address object?
    if isinstance(value, Address):
        return value

    # If we have an integer, assume it is a model primary key.
    elif isinstance(value, int):
        return value

    # A string is considered a raw value.
    elif isinstance(value, str):
        obj = Address(raw=value)
        obj.save()
        return obj

    # A dictionary of named address components.
    elif isinstance(value, dict):

        # Attempt a conversion.
        try:
            return _to_python(value)
        except InconsistentDictError:
            return Address.objects.create(raw=value["raw"])

    # Not in any of the formats I recognise.
    raise ValidationError("Invalid address value.")


##
# A country.
##


class Country(models.Model):
    name = models.CharField(max_length=40, unique=True, blank=True)
    code = models.CharField(max_length=2, blank=True)  # not unique as there are duplicates (IT)

    class Meta:
        verbose_name_plural = "Countries"
        ordering = ("name",)

    def __str__(self):
        return "%s" % (self.name or self.code)


##
# A state. Google refers to this as `administration_level_1`.
##


class State(models.Model):
    name = models.CharField(max_length=165, blank=True)
    code = models.CharField(max_length=8, blank=True)
    country = models.ForeignKey(Country, on_delete=models.CASCADE, related_name="states")

    class Meta:
        unique_together = ("name", "country")
        ordering = ("country", "name")

    def __str__(self):
        txt = self.to_str()
        country = "%s" % self.country
        if country and txt:
            txt += ", "
        txt += country
        return txt

    def to_str(self):
        return "%s" % (self.name or self.code)


##
# A locality (suburb).
##


class Locality(models.Model):
    name = models.CharField(max_length=165, blank=True)
    postal_code = models.CharField(max_length=10, blank=True)
    state = models.ForeignKey(State, on_delete=models.CASCADE, related_name="localities")

    class Meta:
        verbose_name_plural = "Localities"
        unique_together = ("name", "postal_code", "state")
        ordering = ("state", "name")

    def __str__(self):
        txt = "%s" % self.name
        state = self.state.to_str() if self.state else ""
        if txt and state:
            txt += ", "
        txt += state
        if self.postal_code:
            txt += " %s" % self.postal_code
        cntry = "%s" % (self.state.country if self.state and self.state.country else "")
        if cntry:
            txt += ", %s" % cntry
        return txt


##
# An address. If for any reason we are unable to find a matching
# decomposed address we will store the raw address string in `raw`.
##


class Address(models.Model):
    street_number = models.CharField(max_length=20, blank=True)
    route = models.CharField(max_length=100, blank=True)
    locality = models.ForeignKey(
        Locality,
        on_delete=models.CASCADE,
        related_name="addresses",
        blank=True,
        null=True,
    )
    raw = models.CharField(max_length=200)
    formatted = models.CharField(max_length=200, blank=True)
    latitude = models.FloatField(blank=True, null=True)
    longitude = models.FloatField(blank=True, null=True)

    class Meta:
        verbose_name_plural = "Addresses"
        ordering = ("locality", "route", "street_number")

    def __str__(self):
        if self.formatted != "":
            txt = "%s" % self.formatted
        elif self.locality:
            txt = ""
            if self.street_number:
                txt = "%s" % self.street_number
            if self.route:
                if txt:
                    txt += " %s" % self.route
            locality = "%s" % self.locality
            if txt and locality:
                txt += ", "
            txt += locality
        else:
            txt = "%s" % self.raw
        return txt

    def clean(self):
        if not self.raw:
            raise ValidationError("Addresses may not have a blank `raw` field.")

    def as_dict(self):
        ad = dict(
            street_number=self.street_number,
            route=self.route,
            raw=self.raw,
            formatted=self.formatted,
            latitude=self.latitude if self.latitude else "",
            longitude=self.longitude if self.longitude else "",
        )
        if self.locality:
            ad["locality"] = self.locality.name
            ad["postal_code"] = self.locality.postal_code
            if self.locality.state:
                ad["state"] = self.locality.state.name
                ad["state_code"] = self.locality.state.code
                if self.locality.state.country:
                    ad["country"] = self.locality.state.country.name
                    ad["country_code"] = self.locality.state.country.code
        return ad


class AddressDescriptor(ForwardManyToOneDescriptor):
    def __set__(self, inst, value):
        super(AddressDescriptor, self).__set__(inst, to_python(value))


##
# A field for addresses in other models.
##


class AddressField(models.ForeignKey):
    description = "An address"

    def __init__(self, *args, **kwargs):
        kwargs["to"] = "address.Address"
        # The address should be set to null when deleted if the relationship could be null
        default_on_delete = models.SET_NULL if kwargs.get("null", False) else models.CASCADE
        kwargs["on_delete"] = kwargs.get("on_delete", default_on_delete)
        super(AddressField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, virtual_only=False):
        from address.compat import compat_contribute_to_class

        compat_contribute_to_class(self, cls, name, virtual_only)

        setattr(cls, self.name, AddressDescriptor(self))

    def formfield(self, **kwargs):
        from .forms import AddressField as AddressFormField

        defaults = dict(form_class=AddressFormField)
        defaults.update(kwargs)
        return super(AddressField, self).formfield(**defaults)
